################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################
require "rest-client"
require "restclient/request"
require "base64"
require "json"
require "uri"
require "fileutils"

class RAMSession
  
    # Constants
    APPLICATION_JSON = "application/json"
    LANGUAGE_EN = "en"
    QUOTE="\""
    SPACE=" "
      
    @ram_conf
    @ram_headers
    @category_schemas
    @attributes    
    
    # initialization method
    #
    # Parameters
    # ==========
    #   Required
    #      root_server_uri : URL to the ram server. For e.g. http://localhost:8080/ram.ws
    #      user_id : User ID to access ram server
    #      password : password to access RAM server
    #      
    #
    def initialize(root_server_uri, user_id, password, log=nil, log_level=Logger::WARN)
      if log.nil?
          log = Logger.new(STDOUT) 
      end
      @log = log
      @log_level=log_level
      @log.level = log_level   
      RestClient.log = @log if (log_level == Logger::DEBUG)
      
      unless root_server_uri.nil? or user_id.nil? or password.nil?
        @root_server_uri = root_server_uri
        @user_id = user_id
        @password = password
        create_default_header_hash()
        set_user(user_id, password)
      else 
        raise ArgumentError, "RAM requires values for root_server_uri, user_name, and password"
      end
    end
    

    #   Creates an asset from json
    #
    #   Parameters
    #   ==========
    #   asset_as_json : json representing the asset to be created
    #   
    #   Returns
    #   ========
    #   json hash from the response
    #
        
    def create_asset(asset_as_json)
      @log.debug "In create_asset(asset_as_json), asset_as_json: #{asset_as_json.inspect} "
      begin
        response = nil
        resource = 'RAMSecure/oslc/assets'
        url = "#{@root_server_uri}/#{resource}"      
              
        #post JSON, this creates an asset     
        response = RestClient.post url, asset_as_json, @ram_headers
        return JSON.parse(response.body)
      rescue Exception => e
        handle_exception(e)
      end  
    end  
        
    # Gets the asset as json
    #   guid : guid of the asset
    #   version : version of the asset
    #
    def get_asset(guid, version)
      begin
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}"
        response = RestClient.get url, @ram_headers
        return response
      rescue Exception => e
        handle_exception(e)
      end  
    end
    
    # Adds artifact to the asset
    #
    # Parameters
    # ===========
    # guid : GUID of the asset
    # version : Version of the asset
    # file : file to be uploaded as artifact
    # filename : name of the artifact
    #
    # Returns
    # ========
    # url of the artifact
    
    def add_artifact(guid, version, file, filename)  
        if filename.nil?
          filename = File.basename(file)
        end
        filename = URI.escape(filename)
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}/artifacts?name=#{filename}"
        url = "#{@root_server_uri}/#{resource}"
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            args = {
                  :method => "post",
                  :timeout => 90000000,
                  :open_timeout => 90000000,
                  :url => url,
                  :headers => @ram_headers,
                  :payload => File.new(file, 'rb')
             }
             response = RestClient::Request.execute(args)
             #response = RestClient.post url, File.new(file, 'rb'), @ram_headers
            unless response.nil?
              return get_url(guid, version, filename)
            end
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                if (i >= 3)
                   handle_exception(exceptionWithResponse)
                 else
                   #loop for three times : could check for collisionDetectionException in the message
                   if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                     handle_exception(exceptionWithResponse)
                   else
                     @log.info exceptionWithResponse.inspect
                   end
                 end
            rescue Exception => e 
              handle_exception(e)
            end  
        end 
    end 
    
    # Adds artifact to the asset passed as a content
    #
    # Parameters
    # ===========
    # guid : GUID of the asset
    # version : Version of the asset
    # content : content to be uploaded as artifact
    # filename : name of the artifact
    #
    # Returns
    # ========
    # url of the artifact
    
    def add_artifact_content(guid, version, content, filename) 
       
        filename = URI.escape(filename)
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}/artifacts?name=#{filename}"
        url = "#{@root_server_uri}/#{resource}"
        
        #post file into the asset
        headers = @ram_headers.clone
        headers[:content_type] = "text/plain" 
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            response = RestClient.post url, content, headers
            unless response.nil?
              return get_url(guid, version, filename)
            end
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                  if (i >= 3)
                    handle_exception(exceptionWithResponse)
                  else
                    #loop for three times : could check for collisionDetectionException in the message
                    if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                      handle_exception(exceptionWithResponse)
                    else
                      @log.info exceptionWithResponse.inspect
                    end
                  end
            rescue Exception => e 
              handle_exception(e)
            end  
        end   
    end  
    
     
    
    # Downloads the artifact to the specified folder. 
    #
    # guid : GUID of the asset
    # version : version of the asset
    # filename : Artifact name to be downloaded
    # folder : Folder to which the artifact will be downloaded
    #
    def download_artifact(guid, version, filename, folder) 
      begin 
      
        FileUtils.mkdir_p "#{folder}"
        
        filename = File.basename(filename)
         
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}/artifactContents/#{filename}"
        url = "#{@root_server_uri}/#{resource}"
    
        escapedURL = URI.escape(url)
        
        response = nil
        
         args = {
          :method => "get",
          :timeout => 90000000,
          :open_timeout => 90000000,
          :url => escapedURL,
          :headers => @ram_headers
        }
        response = RestClient::Request.execute(args)
        
        unless response.nil?
          file = "#{folder}/#{filename}"
          f = open(file, "wb")    
          f.write(response.body)
          f.close()
          return file
        end
      rescue RestClient::ExceptionWithResponse => exceptionWithResponse
        handle_exception(exceptionWithResponse)
      rescue Exception => e 
        handle_exception(e)
      end  
     end
     
    # Downloads the artifact resource url to the specified folder. 
    #
    # guid : GUID of the asset
    # version : version of the asset
    # resource : Download the resource 
    # folder : Folder to which the artifact will be downloaded
    #
    def download_resource(guid, version, resource, folder) 
      begin 
      
        FileUtils.mkdir_p "#{folder}" 
         
        resource = "RAMSecure/#{resource}"
        url = "#{@root_server_uri}/#{resource}"
        filename = File.basename(resource)
        escapedURL = URI.escape(url)
        
        response = nil
        
         args = {
          :method => "get",
          :timeout => 90000000,
          :open_timeout => 90000000,
          :url => escapedURL,
          :headers => @ram_headers
        }
        response = RestClient::Request.execute(args)
        
        unless response.nil?
          file = "#{folder}/#{filename}"
          f = open(file, "wb")    
          f.write(response.body)
          f.close()
          return file
        end
      rescue RestClient::ExceptionWithResponse => exceptionWithResponse
        handle_exception(exceptionWithResponse)
      rescue Exception => e 
        handle_exception(e)
      end  
     end
     
    # Downloads the asset to the specified folder. 
    #
    # guid : GUID of the asset
    # version : version of the asset
    # folder : Folder to which the artifact will be downloaded
    # filename : Save As filename
    #
    def download_asset(guid, version, folder, filename) 
      begin  

        FileUtils.mkdir_p "#{folder}"
        headers = @ram_headers.clone
        headers[:content_type] =  "application/zip"
         
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}" 
        url = "#{@root_server_uri}/#{resource}.zip"
        
        escapedURL = URI.escape(url)
        if filename.nil?
          filename="#{guid}-#{version}.zip"
        end
        response = nil
        args = {
          :method => "get",
          :timeout => 90000000,
          :open_timeout => 90000000,
          :url => escapedURL,
          :headers => @ram_headers
        }
        response = RestClient::Request.execute(args)
        unless response.nil?
          file = "#{folder}/#{filename}"
          f = open(file, "wb")    
          f.write(response.body)
          f.close()
          return file
        end
        
      rescue RestClient::ExceptionWithResponse => exceptionWithResponse  
        handle_exception(exceptionWithResponse)
      rescue Exception => e 
        handle_exception(e)
      end  
     end
     
    # Gets the url of the asset / artifact
    # If the file is not specified, it will return the url of the asset
    # guid : GUID of the asset
    # version : Version of the asset
    # filename : artifact file name
    
    def get_url(guid, version, filename=nil)
         unless filename.nil? 
           filename = File.basename(filename)
            resource = "oslc/assets/#{guid}/#{version}/artifactContents/#{filename}"
         else
            resource = "oslc/assets/#{guid}/#{version}"
         end
         
         url = "#{@root_server_uri}/#{resource}" 
         escaped_url = URI.escape(url) 
         return escaped_url
    end  
    
    def get_resource_url(resource)
      url = "#{@root_server_uri}/#{resource}" 
      return url
    end
    
    def execute_query(query, asset_properties = nil)
      begin
        resource = "RAMSecure/oslc/simpleQuery"
        url  = "#{@root_server_uri}/#{resource}"
       
        if asset_properties.nil?
          headers = { 
            :params=>{'oslc_asset.query' => query}
          } .merge(@ram_headers)
          response = RestClient.get url, headers 
        else 
          headers = {
            :params=>{ 
              'oslc_asset.query' => query, 
              'oslc_asset.properties' => asset_properties
             }
          } .merge(@ram_headers)
          response = RestClient.get url, headers 
        end
        return response
      rescue Exception => e
        handle_exception(e)
      end 
    end
    
    # Gets the asset as json
    #   resource : resource uri of the asset
    #
    def get_asset_from_url (resource)
      begin
        index = resource.index('RAMSecure')
        if index == -1
           resource = "RAMSecure/#{resource}"
        end
        url = "#{@root_server_uri}/#{resource}"
        response = RestClient.get url, @ram_headers
        return response
      rescue Exception => e
        handle_exception(e)
      end
    end
    
    # Updates the asset with the metadata provided
    # Parameters
    #
    #       guid : Guid of the asset
    #       version : version of the asset
    #       metadata: asset metadata
    
    def update_asset(guid, version, metadata)
      begin
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}"
        
        asset_json = metadata.to_json
        put_response = RestClient.put url, asset_json, @ram_headers
        return put_response
      rescue Exception => e
        handle_exception(e)
      end
    end

    # Deletes the asset with the metadata provided
    # Parameters
    #
    #       guid : Guid of the asset
    #       version : version of the asset
    #       metadata: asset metadata
    
    def delete_asset(guid, version)
      begin
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}"

        delete_response = RestClient.delete url, @ram_headers
        return delete_response
      rescue Exception => e
        handle_exception(e)
      end
    end
  
    # Adds attribute to the asset
    # Parameters
    #       guid : Guid of the asset
    #       version : version of the asset
    #       attribute_json:  attribute json to add
    
    def add_attribute(guid, version, attribute_json) 
        
        # get E-Tag
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}" 
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            response = RestClient.get url, @ram_headers
            asset_data = JSON.parse(response.body)
            
            resource = "RAMSecure/oslc/assets/#{guid}/#{version}?oslc_asset.properties=ram:attributeValue"
            put_url = "#{@root_server_uri}/#{resource}"
            
            attribute_data = JSON.parse(attribute_json)
            attribute_data = {
              "etag" => asset_data["etag"]
            }.merge(attribute_data)
            unless asset_data["ramAttributeValues"].nil?
                #remove the asset attributes
                attributes = asset_data["ramAttributeValues"]
                attributes.each do |attribute| 
                  if ((!attribute["attribute"]["value"].nil?) && (!attribute["attribute"]["value"]["name"].nil?) &&  (attribute["attribute"]["value"]["name"] == attribute_data["ramAttributeValues"][0]["attribute"]["value"]["name"]))
                       asset_data["ramAttributeValues"].delete(attribute)
                  elsif ((!attribute["attribute"]["title"].nil?) && (!attribute["attribute"]["title"]["content"].nil?) && (attribute["attribute"]["title"]["content"][0]== attribute_data["ramAttributeValues"][0]["attribute"]["value"]["name"]))
                        asset_data["ramAttributeValues"].delete(attribute)
                  end
                end
                attribute_data["ramAttributeValues"] = asset_data["ramAttributeValues"] | attribute_data["ramAttributeValues"]
            end
             
            attribute_json_updated = attribute_data.to_json 
            #post attribute into the asset 
            put_response = RestClient.put put_url, attribute_json_updated, @ram_headers
            return put_response
            
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                if (i >= 3)
                  handle_exception(exceptionWithResponse)
                else
                  #loop for three times : could check for collisionDetectionException in the message
                  if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                    handle_exception(exceptionWithResponse)
                  else
                    @log.info exceptionWithResponse.inspect
                  end
                end
            rescue Exception => e 
              handle_exception(e)
            end  
        end   
    end
    
    # Adds attribute to the asset
    # Parameters
    #       guid : Guid of the asset
    #       version : version of the asset
    #       attribute_json:  attribute json to add
    
    def add_artifact_link(guid, version, artifact_json)
        # get E-Tag
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}" 
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            response = RestClient.get url, @ram_headers
            asset_data = JSON.parse(response.body)
            
            resource = "RAMSecure/oslc/assets/#{guid}/#{version}?oslc_asset.properties=artifact"
            put_url = "#{@root_server_uri}/#{resource}"
            
            artifact_data = JSON.parse(artifact_json)
            artifact_data = {
              "etag" => asset_data["etag"]
            }.merge(artifact_data)
            unless asset_data["artifacts"].nil?
                artifact_data["artifacts"] = asset_data["artifacts"] | artifact_data["artifacts"]
            end
             
            artifact_json_updated = artifact_data.to_json 
            #post attribute into the asset 
            put_response = RestClient.put put_url, artifact_json_updated, @ram_headers 
            return put_response
        
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                if (i >= 3)
                  handle_exception(exceptionWithResponse)
                else
                  #loop for three times : could check for collisionDetectionException in the message
                  if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                    handle_exception(exceptionWithResponse)
                  else
                    @log.info exceptionWithResponse.inspect
                  end
                end
            rescue Exception => e 
              handle_exception(e)
            end  
        end   
    end
    
    # Adds relationship to the asset
    # Parameters
    #       guid : Guid of the asset
    #       version : version of the asset
    #       relationship_json:  relationship json to add
    
    def add_relationship(guid, version, relationship_json) 
        
        # get E-Tag
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}" 
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            response = RestClient.get url, @ram_headers
            asset_data = JSON.parse(response.body)
            
            resource = "RAMSecure/oslc/assets/#{guid}/#{version}?oslc_asset.properties=oslc_asset:relation"
            put_url = "#{@root_server_uri}/#{resource}"
            
            relationship_data = JSON.parse(relationship_json)
            relationship_data = {
              "etag" => asset_data["etag"]
            }.merge(relationship_data)
            unless asset_data["relationships"].nil?
                relationship_data["relationships"] = asset_data["relationships"] | relationship_data["relationships"]
            end
             
            relationship_json_updated = relationship_data.to_json 
            #post attribute into the asset 
            put_response = RestClient.put put_url, relationship_json_updated, @ram_headers
            return put_response
        
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                if (i >= 3)
                  handle_exception(exceptionWithResponse)
                else
                  #loop for three times : could check for collisionDetectionException in the message
                  if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                    handle_exception(exceptionWithResponse)
                  else
                    @log.info exceptionWithResponse.inspect
                  end
                end
            rescue Exception => e 
                handle_exception(e)
            end  
        end   
    end
    
    # Get all the RAM category schemas
    def get_category_schemas()
      if @category_schemas.nil?
        resource = "internal/categorySchemas"
        url = "#{@root_server_uri}/#{resource}"
      
        response = RestClient.get url, @ram_headers
        @category_schemas = JSON.parse(response.body)
      end
      return @category_schemas
    end
    
    # Get category schema by name
    def get_category_schema(name)
      schemas = get_category_schemas()
      unless schemas.nil?
          schemas.each do |schema|
               if schema["name"]==name
                 resource = schema["href"]
                 url = "#{@root_server_uri}/#{resource}"
                 response = RestClient.get url, @ram_headers
                 return JSON.parse(response.body)
               end
          end
      end
      return nil
    end 
    
    # Adds relationship to the asset
    # Parameters
    #       guid : Guid of the asset
    #       version : version of the asset
    #       category_json:  relationship json to add
    
    def add_category(guid, version, category_json, parent_category=nil, remove_exclusiveness=true) 
        
        # get E-Tag
        resource = "RAMSecure/oslc/assets/#{guid}/#{version}"
        url = "#{@root_server_uri}/#{resource}" 
        
        i=0
        while i<=3 #loop for 2 times for collision detection exception
            begin 
            i +=1
            #post file into the asset
            response = RestClient.get url, @ram_headers
            asset_data = JSON.parse(response.body)
            
            resource = "RAMSecure/oslc/assets/#{guid}/#{version}?oslc_asset.properties=ram:category"
            put_url = "#{@root_server_uri}/#{resource}"
            
            asset_data=remove_exclusive(parent_category,asset_data, remove_exclusiveness)
            
            category_data = JSON.parse(category_json)
            category_data = {
              "etag" => asset_data["etag"]
            }.merge(category_data)
            unless asset_data["ramCategories"].nil?
                category_data["ramCategories"] = asset_data["ramCategories"] | category_data["ramCategories"]
            end
             
            category_json_updated = category_data.to_json 
            #post attribute into the asset 
            put_response = RestClient.put put_url, category_json_updated, @ram_headers
            return put_response
        
            
            rescue RestClient::ExceptionWithResponse => exceptionWithResponse
                if (i >= 3)
                  handle_exception(exceptionWithResponse)
                else
                  #loop for three times : could check for collisionDetectionException in the message
                  if !exceptionWithResponse.instance_of?(RestClient::InternalServerError)
                    handle_exception(exceptionWithResponse)
                  else
                    @log.info exceptionWithResponse.inspect
                  end
                end
            rescue Exception => e 
              handle_exception(e)
            end  
        end   
    end
    
    def get_attributes ()
      if @attributes.nil?
        resource = "RAMSecure/oslc/attributes"
        url = "#{@root_server_uri}/#{resource}" 
        response = RestClient.get url, @ram_headers
        @attributes = JSON.parse(response.body)
      end
      return @attributes
    end  
      
     # Get attribute by name
     def get_attribute(name)
        attributes = get_attributes()
        unless attributes.nil?
          attributes.each do |attribute|
                 if attribute["name"]==name
                    return attribute
                 end
          end
        end
        return nil
     end 
     
     #get user by Id
     def get_user(id)
       begin
         resource = "RAMSecure/oslc/users/#{id}"
         url = "#{@root_server_uri}/#{resource}" 
         response = RestClient.get url, @ram_headers
         
         user = JSON.parse(response.body)
         return user
       rescue Exception => e
         @log.debug e.inspect
       end  
       return nil
     end 
    
   
  
   private
    ################################################################################
    # Private Methods                                                              #
    ################################################################################
    
   
    
     def set_user(user_id, password)
        clear_http_session()
        set_authorization_header(user_id, password)
     end
       
    # HTTP Header Manipulation
    
    def create_default_header_hash(headers={})
      @ram_headers = {
        :accept => APPLICATION_JSON,
        :accept_language => LANGUAGE_EN,
        :content_type => APPLICATION_JSON
      }.merge(headers)
    end
    
    def set_http_session(session_cookies)
      @ram_headers[:cookies] = session_cookies
      @ram_headers.delete(:authorization)
    end
    
    def clear_http_session
      @ram_headers.delete(:cookies)
    end
    
    def set_authorization_header(user_id, password)
      auth_value = Base64.encode64("#{user_id}:#{password}")
      @ram_headers[:authorization] = "Basic #{auth_value}"
    end 
    
    def handle_exception(ex)
      @log.info  "Library server on #{@root_server_uri} using user account #{@user_id}/****"
      if ex.instance_of?(Errno::ETIMEDOUT)
        @log.info "Please verify the url to the library server is correct."
      elsif ex.instance_of?(Errno::EHOSTUNREACH)
        @log.info "Please verify that the hostname is reachable from the client. Verify that the Admin > Configuration > 'Web Server Path' and 'Web Services Path' are configured correctly."
      elsif ex.instance_of?(RestClient::InternalServerError)  
        @log.info "An error occurred while communicating with the Library Server. The server debug logs available under Admin > Tools > Ram Debug Logs may help you diagnose the cause of the failure."
      elsif ex.instance_of?(RestClient::ResourceNotFound)
        @log.info "An attempt to access a resource which does not exist has failed. Please verify that the resource exists."
      elsif ex.instance_of?(RestClient::Unauthorized)   
        @log.info "Please verify the url, user and password to the library server are correct."
      end
      @log.info ex.inspect
      @log.info "Set the logger level to Debug for more information." if !@log.debug?
      raise ex
    end
    
    def remove_exclusive(parent_category,asset_data,remove_exclusivness=true)
      asset_categories=asset_data["ramCategories"]
      if !parent_category.nil? && !asset_categories.nil?
         if parent_category["exclusive"]
            subCategories = parent_category["subCategory"]
            subCategories.each do |subCategory|
                #if the sub-category has exclusive children remove the children
                asset_categories.each_with_index { |asset_category, i|
                     if (!remove_exclusivness && asset_category['value']==subCategory['displayPath'])
                       raise "Cannot re-categorize the asset (\'#{asset_data["title"]}, id=#{asset_data["id"]["ras:GUID"]}, version=#{asset_data["id"]["ras:version"]}\'), already categorized as #{asset_category['value']}"
                     else
                       asset_categories.delete_at(i) if asset_category['value']==subCategory['displayPath']
                     end
                 } 
                 #raise "Cannot re-categorize the asset, already categorized as #{asset_category['value']}" if (!remove_exclusivness && asset_category['value']==subCategory['displayPath'])
                 #asset_categories.delete_at(i) if asset_category['value']==subCategory['displayPath']
              
            end # end of each loop
          end
      end
      return asset_data
    end
    
    
    def getfileNameFrom_ContentDisposition(response)
      headers = response.headers
    
      contentDisposition = headers[:content_disposition]


      unless  contentDisposition.nil?  # if not nil
        splitted_str = contentDisposition.split(";")
      
        for name in splitted_str
          isFileName = name.start_with?(" filename") #there is a space too
          if (isFileName)
             fileNameString = name.split("=")
             fileNameValue = fileNameString[1]
          end  
        end  
      end
     
      #remove quotes
      size = fileNameValue.size
      fileNameValue = fileNameValue[1..size-2]
      return fileNameValue
    end  
    
    
end


