require 'timeout'
require 'weaver/cli/weaver_cli_plugin'
require 'weaver/cli/plugins/env_cli_plugin.rb'
require 'weaver/deployers/plugins/iwd_deployer'
require 'weaver/generators/inf_facade_generator'
require 'weaver/iwd/iwd_webmock_generator'
require 'weaver/core/weaver_transforms'
require 'weaver/iwd/iwd_weaver_inf_generator'
require 'weaver/core/constants'
require 'weaver/dsl/weaver_parser'
require 'weaver/environment'
require 'weaver/core/logging'
require 'weaver/iwd/iwd_script_generator'
require 'zip/zip'
require 'zip/zipfilesystem'
require 'devops_deployment_lib/providers/iwd_virtual_system_client'
require 'weaver/iwd/iwd_script_generator'

module Weaver
  module Cli
    module Plugins
      # Defines plugins used by Weaver's command-line support for {IBM Workload Deployer}[http://www-01.ibm.com/software/webservers/workload-deployer/].
      module Iwd
        # IWD plugin for the Weaver CLI, allows to import an IWD pattern and deploy a Weaver
        # BOM to IWD.  Can also generate script packages.
        #
        # Syntax: weaver iwd [task] [options]
        #
        # Author:: Florian Rosenberg, Ann Marie Fred
        #
        class IwdCliPlugin < Weaver::Cli::WeaverCliPlugin

          # Special non-Thor extension to provide help for the arguments of the commands
          ARG_DEFS = {
            'PATTERN_ID' => 'The ID number of the pattern, which is found at the end of the URL of the pattern in the web client.',
            'ENVIRONMENT_FILE' => 'A .weaver file defining an environment',
            'INFRASTRUCTURE_FILE' => 'A .weaver file defining an infrastructure', 
            'FACADE_NAME' => 'The name to give the generated Weaver facade'
          }
        
          namespace :iwd
          namespace_description "IBM Workload deployer (IWD) related tasks"

          desc "import PATTERN_ID", "Import an IWD pattern as infrastructure definition."
          method_option :host, :type => :string, :aliases => "-h", :desc => "IWD hostname."
          method_option :user, :type => :string, :aliases => "-u", :desc => "IWD username."
          method_option :pwd, :type => :string, :aliases => "-p", :desc => "IWD password."
          method_option :nossl, :type => :boolean, :desc => "Disable SSL certificate validation."
          method_option :loc, :type => :string, :aliases => "-l", :desc => "The location where to put the file."
          method_option :skip_topologies, :type => :boolean, :aliases => '-s', :desc => "Skip generation of ICCT virtual image topology information."
          method_option :with_yaml, :type => :boolean, :aliases => '-y', :desc => "Write the infrastructure with mutable properties in a yml file."
          # Import an IWD Pattern and produce a Weaver document that models it
          def import(pattern_id = nil)
            collect_properties
            setup_logging

            # This is a bug in Thor and JRuby so we have to handle that manually
            raise Thor::Error.new(%Q{"import" was called incorrectly. Call as "import PATTERN_ID".}) unless pattern_id
            raise Thor::Error.new(%Q{"import" was called incorrectly. PATTERN_ID must be a number.}) unless pattern_id.to_i >= 1
            raise Thor::Error.new(%Q{"import" was called incorrectly. PATTERN_ID must be a number.}) unless pattern_id.to_i.to_s == pattern_id

            host, username, password = process_iwd_connection_info()            
            
            # create a infrastructure generator for IWD
            generator = Weaver::Iwd::IwdWeaverInfGenerator.new(host, username, password, !options[:nossl]) { |lvl, status, msg| weaver_report(lvl, status, msg) }              
            inf = invoke_iwd_command(host, username, password, pattern_id) { generator.generate_infrastructure(pattern_id, !options[:skip_topologies]) }   
            
            # generate infrastructure definition filename and target directory
            full_filename = filename = inf.id.to_s + Weaver::Core::WEAVER_SUFFIX

            dir = options[:loc] || '.'
            if options[:loc]
              full_filename = File.join(options[:loc], filename)
            end
            with_yaml = ""
            yml_full_filename = ""
            yml_filename = ""
            if options[:with_yaml]
              yml_full_filename = full_filename.chomp(File.extname(full_filename)) + '.yml'
              yml_filename = File.basename(yml_full_filename)
              with_yaml = "with yml '#{yml_filename}' "
            end
            say_status("Saving", "Weaver topology '#{filename}' #{with_yaml}in '#{dir}'", :green)

            FileUtils.mkdir_p(File.dirname(filename)) unless File.directory?(File.dirname(filename))
              
            # creating a backup copy if file already exists
            if File.exists?(full_filename)
              say_status("Creating", "File '#{filename}' already exists. Creating backup '#{filename}.saved'", :green)
              FileUtils.copy(full_filename, "#{full_filename}.saved")
            end
            if !yml_full_filename.nil? && File.exists?(yml_full_filename)
              say_status("Creating", "File '#{yml_filename}' already exists. Creating backup '#{yml_filename}.saved'", :green)
              FileUtils.copy(yml_full_filename, "#{yml_full_filename}.saved")
            end

            begin
              if yml_full_filename.nil?
                generator.save_infrastructure(inf, full_filename)
              else
                generator.save_infrastructure(inf, full_filename, yml_full_filename)
              end
            rescue SystemCallError => se
              # For example, Errno::EACCES: Permission denied
              say_status("Error", se.message, :red)
              exit(1)
            end
          end

          desc "webmock PATTERN_ID", "Generate a set of WebMock files for an  IWD pattern."
          method_option :host, :type => :string, :aliases => "-h", :desc => "IWD hostname."
          method_option :user, :type => :string, :aliases => "-u", :desc => "IWD username."
          method_option :pwd, :type => :string, :aliases => "-p", :desc => "IWD password."
          method_option :nossl, :type => :boolean, :desc => "Disable SSL certificate validation."
          method_option :loc, :type => :string, :aliases => "-l", :desc => "The location where to put the files."
          # Create files that mock IWD HTTP REST interaction for use with http://rubygems.org/gems/webmock
          def webmock(pattern_id = nil)
            collect_properties
            setup_logging
           
            # This is a bug in Thor and JRuby so we have to handle that manually
            raise Thor::Error.new(%Q{"webmock" was called incorrectly. Call as "webmock PATTERN_ID".}) unless pattern_id
            
            host, username, password = process_iwd_connection_info()            

            # create a infrastructure generator for IWD
            dir = options[:loc] || '.'
            say_status("Saving", "WebMock files in '#{dir}'", :green)
            
            generator = Weaver::Iwd::IwdWeaverWebMockGenerator.new(host, username, password, !options[:nossl])              
            invoke_iwd_command(host, username, password, pattern_id) { generator.generate_webmocks(pattern_id, options[:loc]) }   
          end
          
          desc "create_script_package", "Create a script package that can be executed within a virtual system pattern."
          method_option :recipe_url, :type => :string, :required => false, :desc => "URL of cookbook containing recipe."
          method_option :recipe, :type => :string, :required => false, :desc => "Recipe to covert to a bundle."
          method_option :chef_path, :type => :string, :required => false, :desc => "Location of Chef cookbooks."
          method_option :role, :type => :string, :required => false, :desc => "Chef role to covert to a script package."
          method_option :automation, :type => :string, :required => false, :desc => "Automation model (generated by Weaver)."
          method_option :host, :type => :string, :required => false, :desc => "URL of cloud provider instance."
          method_option :user, :type => :string, :required => false, :desc => "User of specified cloud provider instance."
          method_option :pwd, :type => :string, :required => false, :desc => "Password for specified user of specified cloud provider instance."
          method_option :nossl, :type => :boolean, :desc => "Disable SSL certificate validation."
          def create_script_pkg(name, version, description = nil)
            collect_properties
            setup_logging
  
            # This is a bug in Thor and JRuby so we have to handle that manually
            raise Thor::Error.new(%Q{"create_script_package" was called incorrectly. name required.}) unless name
            raise Thor::Error.new(%Q{"create_script_package" was called incorrectly. version requried.}) unless version
            
            chef_path, role_file = identify_recipe()
            @@logger.debug("Processing '#{chef_path}/roles/#{role_file}' to create script package '#{name}' with version '#{version}'")
            
            @@logger.debug("chef automation file is: #{options[:automation]}")
            
            if options[:automation]
               model = Weaver::Dsl::WeaverParser.new().parse(File.new(options[:automation]))
            else
               inspector = Weaver::Chef::ChefNodeInspector.new(chef_path.split(File::PATH_SEPARATOR)) do |level, msg|
                 report_message("Importing", level, msg)
               end # end do
               model = inspector.model_role(role_file, ".")
            end # end if options[:automation]
            
            # identify/create files to include in run/config operation for script package
            cookbook_archive = archive_cookbooks(chef_path)
            
          	generator = Weaver::Iwd::IwdScriptGenerator.new({'name' =>name, 'version' => version, 'description' => description, 'role_file' => role_file, 'env_model' => model})
            metadata = create_metadata(generator.to_json(), chef_path)
            run_script = create_run_script(model)
            
            say_status("Creating", "Creating script package with archive: '#{cookbook_archive}' and '#{run_script}' and '#{metadata}'", :green)
			script_package = create_script_package_archive(name, version, cookbook_archive, run_script, metadata, options[:automation])
			
            say_status("Saving", "Saved script package '#{script_package}'", :green)
            
            # delete temp files
            cleanup_script_package_archive(cookbook_archive, metadata)
            
            # identify parameters for connection for IWD
            host = options[:host]
            username = options[:user] 
            password = options[:pwd]
            if (host && username && password)        
            	@@logger.debug("Identified cloud provider connection details as: #{username}@#{host}")
            	client_config = {
                	:log_level => Logger::INFO,
                	:log_file => STDOUT
            	}
            	# create script package in IWD
            	client = DevopsDeploymentLib::IwdVirtualSystemClient.new(host, username, password, client_config)
            	if options[:nossl]
            		client.ssl_verify_mode=OpenSSL::SSL::VERIFY_NONE 
            	end
            	invoke_iwd_command(host, username, password, "") { client.create_script_package(name, script_package, true) }   
            	say_status("Uploading", "Uploaded script package '#{script_package}' to the cloud provider", :green)
  			end
  
            return script_package
          end # end create_script_package method
          

          no_tasks do
            # Validate IWD credentials
            def process_iwd_connection_info()
              host = options[:host]
              say_status("Error", "Argument 'host' not set.", :red) && exit(1) unless host
              username = options[:user] 
              say_status("Error", "Argument 'user' not set.", :red) && exit(1) unless username
              password = options[:pwd]
              say_status("Error", "Argument 'pwd' not set.", :red) && exit(1) unless password
              return host, username, password
            end
            
            # Invoke an IWD command
            def invoke_iwd_command(host, username, password, pattern_id, &cmd)
              begin
                cmd.call()
              rescue OpenSSL::SSL::SSLError => se
                say_status("Error", "SSL certificate verification failed. Please run with --nossl. Original message: '#{se.to_s}'", :red)
                exit(1)
              rescue => re
                if re.to_s =~ /404/
                  say_status("Error", "Pattern with id '#{pattern_id}' does not exist. Original message: '#{re.to_s}'", :red)
                  exit(1)
                elsif re.to_s =~ /401/
                  say_status("Error", "User '#{username}' is unauthorized. Original message: '#{re.to_s}'", :red)
                  exit(1)
                elsif re.to_s =~ /403/
                  say_status("Error", "User '#{username}' is unauthorized. Original message: '#{re.to_s}'", :red)
                  exit(1)                
                else
                  say_status("Error", "Unknown problem, please check the server connectivity. Original message: '#{re.to_s}'", :red)
                  exit(1)
                end
              rescue SocketError => soe
                say_status("Error", "Unknown hostname '#{host}'. Original message: '#{soe.to_s}'", :red)
                exit(1)
              end
            end    
            
            # Identify a chef recipe to convert
            def identify_recipe()
               @@logger.debug("Identifying recipe to convert")
               # 'chef_path' is a path to a directory containing cookbooks, roles and nodes
               # same as 'file_cache_path' in a chef-solo config file
               chef_path = options[:chef_path] || ENV['CHEF_SETUP_PATHS']
               @@logger.debug("Chef setup claimed to be in: #{chef_path}")
               validate_path(chef_path) unless chef_path.nil?
            
               # download cookbook if needed
               url = options[:recipe_url]
               @@logger.debug("Need to download cookbook '#{url}'") if url
               chef_path = fetch_cookbook(url, chef_path) if url
               say_status("Error", "Parameter 'path' (or environment variable 'CHEF_SETUP_PATHS') not set.", :red) && exit(1) unless chef_path
               @@logger.debug("Chef setup confirmed to be in: #{chef_path}")
               
               # chef_path should be defined and cookbook should be in chefpath/cookbooks
               cookbook, recipe = process_recipe_info()
               @@logger.debug("Will convert recipe: #{cookbook}::#{recipe}") unless cookbook.nil?
               
               role = options[:role] || (create_role(cookbook, recipe, chef_path) unless cookbook.nil?)
               @@logger.debug("Converting role: '#{role}'") unless role.nil?
               say_status("Error", "No role identified for conversion to script package.  Specify the role file name using the --role option.", :red) && exit(1) unless role
               
               return chef_path, role                
            end # end identify_recipe method
            
            # Fetch a remote cookbook to a local file
            def fetch_cookbook(url, path)
               @@logger.debug("Fetching cookbook '#{url}' to '#{path}'")
               # create_chef_files if path.nil?
               return path
            end # end fetch_cookbook method
            
            # Validate a path
            def validate_path(path)
               @@logger.debug("validating path '#{path}'")
               unless path.nil?
                  path.split(File::PATH_SEPARATOR).each do |dirname|
                     @@logger.debug("validating sub element '#{dirname}'")
                     unless File.directory?(dirname)
                        say_status("Error", "Directory #{dirname} does not exist", :red) && exit(1)
                     else
                        say_status("Error", "#{dirname} is not a Chef directory (ie, does not contain 'cookbooks'; or 'roles' subdirectories.", :red) && exit(1) unless File.directory?(File.expand_path('cookbooks', dirname)) || File.directory?(File.expand_path('roles', dirname))
                     end # end unless File.directory?(dirname)
                  end # each do
               end # end unless path.nil?
            end # end validate_path method
            
            # Find a recipe to convert
            def process_recipe_info()
               @@logger.debug("Looking for recipe to convert")
               recipe = options[:recipe]
               unless recipe.nil?
                  cr = recipe.split("::")
                  cookbook = cr[0];
                  if (1 == cr.length) 
                     recipe = "default"
                  else 
                     recipe = cr[1]
                  end # if (1 == cr.length)
               end # end unless recipe.nil?
               
               return cookbook, recipe
            end # end process_recipe_info method
            
            # Create a chef role file
            def create_role(cookbook, recipe, chef_path)
               @@logger.debug("Creating role for '#{cookbook}::#{recipe} in #{chef_path}'")
               role = cookbook + "_" + recipe
               role_filepath = "#{chef_path}/roles/" + role + "_role.rb"
               file = File.open(role_filepath, 'w')
               file.write(%Q{name "#{cookbook}"\n})
               file.write(%Q{description "Generated role for deploying #{cookbook}::#{recipe}"\n})
               file.write(%Q{run_list "recipe[#{cookbook}::#{recipe}]"})
               return role
            end # end create_role method
            
            # Create a cbscript.json file
            def create_metadata(json, path)
               @@logger.debug("Creating cbscript.json")
               filepath = "cbscript.json"
               file = File.open(filepath, 'w')
               file.write(json)
               file.write("\n")
               file.close
               return filepath
            end # end create_metadata method
            
            
            # delete temp files
            def cleanup_script_package_archive(cookbook_archive, metadata)
               @@logger.debug("Deleting #{cookbook_archive} and #{metadata}")
               File.delete(cookbook_archive)
               File.delete(metadata)
            end #end cleanup_script_package_archive
            
          # @private
          # :stopdoc:
          # Report a status message running a command
          def report_message(status, level, msg)
            if level == Logger::ERROR || level == Logger::FATAL
              say_status(status, msg, :red)
            elsif level == Logger::WARN
              say_status(status, msg, :yellow)
            else
              say_status(status, msg, :green)
            end
          end
          # :startdoc:
            
           # Pack chef cookbooks into a .tgz file
           def archive_cookbooks(chef_path)
              cookbook_dir = chef_path
              cookbook_archive = "cookbooks.tgz"
              # cookbook_dir = find_cookbook(chef_path)
              `tar -czf #{cookbook_archive} -C #{cookbook_dir} .`
              return cookbook_archive
           end # end archive_cookbooks method
           
           # Create the Chef bootstrapping script for this recipe
           def create_run_script(model)
              return  File.join(File.dirname(__FILE__), "..", "..", "iwd", 'bin', 'chef_script_run.sh')
           end # end create_run_script method
              
           # Pack up a script package into a .zip file
           def create_script_package_archive(name, version, cookbook_archive, run_script, metadata, automation)
              script_package_archive = "#{name}.zip"
              FileUtils.rm script_package_archive, :force=>true
              zipfile = Zip::ZipFile.open(script_package_archive, 'wb')
              cookbookfile = File.open(cookbook_archive, 'r')
              zipfile.add(File.basename(cookbook_archive), cookbookfile)
              scriptfile = File.open(run_script, 'r')
              zipfile.add(File.basename(run_script), scriptfile)
              metafile = File.open(metadata, 'r')
              zipfile.add(File.basename(metadata), metafile)
              metafile = File.open(automation, 'r')
              zipfile.add("role.weaver", metafile)
              zipfile.close()
              return script_package_archive
           end # end create_script_package method        
            
            # Send a message to the user
            def weaver_report(lvl, status, msg)
              
              if !options[:no_color]
                colors = { Logger::INFO => :green, Logger::WARN => :yellow, Logger::ERROR => :red, Logger::FATAL => :red, Logger::DEBUG => :clear }
              else
                colors = { }
              end
              
              # Ignore DEBUG messages, we don't want to actually display them, only :abort them on request from a GUI
              say_status(status, msg, colors[lvl]) unless lvl == Logger::DEBUG
            end
                      
            # @private
            # :stopdoc:
            # Check that +raw_renamings+ are valid for pattern
            def unused_raw_renamings(entity, raw_renamings)
              used = Set.new
              used_raw_renamings(entity, raw_renamings, used)
              return Set.new(raw_renamings).subtract(used)
            end
            
            # @private
            def used_raw_renamings(entity, raw_renamings, used_raw_renamings)
              entity.child_entities.each do |id, child|
                if raw_renamings.has_key?(id)
                  used_raw_renamings.add([ id, raw_renamings[id] ])
                end
                
                used_raw_renamings(child, raw_renamings, used_raw_renamings)
              end
            end
            
          end # end of no_tasks
          # :startdoc:

          desc "facade INFRASTRUCTURE_FILE FACADE_NAME", "Create a facade infrastructure file."
          method_option :loc, :type => :string, :aliases => "-l", :desc => "Location of the generated .weaver file."
          method_option :renamings, :type => :string, :aliases => "-r", :desc => "Comma-separated name=value pairs, 'name' is the after, 'value' is the before."
          # Create a Weaver document that represents an IWD Pattern, possibly with different IDs, similar to the
          # 'Facade' design pattern.
          def facade(infrastructure_file = nil, facade_name = nil)
            collect_properties
            setup_logging            

            # This is a bug in Thor and JRuby so we have to handle that manually
            raise Thor::Error.new(%Q{"facade" was called incorrectly. Call as "facade INFRASTRUCTURE_FILE FACADE_NAME".}) unless infrastructure_file
            raise Thor::Error.new(%Q{"facade" was called incorrectly. Call as "facade INFRASTRUCTURE_FILE FACADE_NAME".}) unless facade_name

            validate_file_exist?(infrastructure_file)
            
            full_filename = filename = facade_name + WEAVER_SUFFIX
            dir = options[:loc] || '.'
            if options[:loc]
              full_filename = File.join(options[:loc], filename)
            end
            say_status("Saving:", "Infrastructure definition '#{filename}' in '#{dir}'", :green)
            
            # parse renamings
            raw_renamings = {}
            unless options[:renamings].nil?
              options[:renamings].split(',').each do |item|
                if item.include?('=')
                  value = item.split('=')
                  raw_renamings[value[1].to_sym] = value[0].to_sym
                else
                  say_status("Error:", "Unable to parse 'renamings'. No separator '=' found in '#{item}', deploy_params='#{options[:renamings]}'.", :red)
                  exit
                end
              end
            end
            
            wp = Weaver::Dsl::WeaverParser.new()
            root = wp.parse(infrastructure_file)
            inf = root.infrastructure
            
            unused_renamings = unused_raw_renamings(inf, raw_renamings)
            unused_renamings.each do |renaming|
              say_status("Warn:", "ignoring #{renaming[1]}=#{renaming[0]} because #{renaming[0].inspect} doesn't occur in Infrastructure")
            end
            
            inf_path = File.expand_path(infrastructure_file)
            gen = Weaver::Generators::InfFacadeGenerator.new(inf, inf_path)
            gen.ensure_all_nodes_renamed(inf, raw_renamings)
            renamings_tree_form = gen.build_tree_form(inf, raw_renamings)
            
            FileUtils.mkdir_p(File.dirname(filename)) unless File.directory?(File.dirname(filename))
              
            # creating a backup copy if file already exists
            if File.exists?(full_filename)
              say_status("Creating:", "File '#{filename}' already exists. Creating backup '#{filename}.saved'", :green)
              FileUtils.copy(full_filename, "#{full_filename}.saved")
            end
            
            facade = gen.generate_inf_facade(facade_name.symbolize, full_filename, renamings_tree_form)
             
          end
        end
      end

    end
  end
end