require 'weaver/core/base'
require 'weaver/utils/indent'

module Weaver
  module Core
    # A mixin for classes such as Link which may contain Link
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module LinkContainerMixin
      include Weaver::Core
      
      # Let all classes that include LinkContainerMixin have a method to return their links
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Link
          def links
            # Since mixins have no initialize() to set up @links, initialize @links lazily
            @links ||= {} unless frozen?
            return @links || {}
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_links)
        baseclazz.register_validate_handler(:validate_links)
        baseclazz.register_child_enumerator(:links)           
        baseclazz.register_reference_freshener(:link_freshener)
      end
      
      # Deeply copy @links
      def initialize_copy_links(other_link_container)
        # At this point, self.initialize_copy should have cloned @links but the values are shallow.
        # However at this point 'nodes' should be deep
        
        @links = {}
        other_link_container.links.each do |id, link|
          dup = link.clone()
          dup.scope = self
          # The link source and target must be handled later, by link_freshener().
          @links[id] = dup
        end
      end
      
      # Validate @links
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_links(out_messages, desired_phases = nil)
        links.each { |id, link| link.validate_collect(out_messages, desired_phases) }
      end
      
      # Finish the work started by initialize_copy_links() and fix up the source/target
      def link_freshener(old_to_new)
        links.each do |id, link|
          link.source = old_to_new[link.source] if old_to_new.key?(link.source) 
          link.target = old_to_new[link.target] if old_to_new.key?(link.target)
        end
      end
      
      protected
      
      # Return a String containing a Weaver description of the links
      def to_weaver_links
        # derive template name according to convention
        template_file = 'data/templates/weaver_core_linkcontainer.erb'
        file = File.join(File.dirname(__FILE__), "..", "..", "..", template_file)
        unless File.exists?(file)
          raise WeaverError, "Cannot find Erubis template file #{template_file} for mixin LinkContainerMixin."
        end
        return Weaver::Utils::Indent.fixup(Erubis::Eruby.new(File.read(file)).result(binding))
      end
    end
    
  end
end