require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Topology which may contain Providers
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module ProviderContainerMixin
      # Let all classes that include TopologyContainerMixin have a method to return their topologies
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Topology
          def providers
            # Since mixins have no initialize() to set up @topologies, initialize @topologies lazily
            @providers ||= {} unless frozen?
            return @providers || {}
          end
          
          # List of mixes attached
          def provider_mixins
            @provider_mixins ||= [] unless frozen?
            return @provider_mixins
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_providers)
        baseclazz.register_validate_handler(:validate_providers)
        baseclazz.register_child_enumerator(:providers)           
        
      end
      
      # Deeply copy @topologies
      def initialize_copy_providers(other_provider_container)
      
        # At this point, self.initialize_copy should have cloned @providers but the values are shallow.
        @providers = {}
        other_provider_container.providers.each do |id, provider|
          raise WeaverError.new("provider (#{id.inspect}) is nil") if provider.nil?
          dup = provider.clone()
          dup.scope = self
          @providers[id] = dup
        end
      end

      # Validate @providers      
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_providers(out_messages, desired_phases = nil)
        providers.each { |id, provider| provider.validate_collect(out_messages, desired_phases) }
      end
    end
    
  end
end