module Weaver
  module Dsl
    # Mix in ComponentContainer in statements that may contain {ComponentStatement}
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    module ComponentContainer
      include Weaver::Core
      
      # * *Args* :
      #   -+base+ the AbstractStatement being mixed with
      def self.included(base)
        base.register_statement_handler([:component_handler])
      end

      # * *Args* :
      #   -+sym_or_model+ <Symbol representing Weaver type, e.g. <code>:software_product</code> | Core::Component
      #   -+id+ Symbol naming entity
      #   -+block+ optional block of additional metadata
      def component_handler(sym_or_model, id, &block)
        return nil unless is_a?(Weaver::Dsl::ComponentContainer) && (sym_or_model.kind_of?(Component) || root_statement.model.weaver_extended_type?(sym_or_model, Component))
        return nil unless id.kind_of?(Symbol) 
        
        prototype = sym_or_model.kind_of?(Symbol) ? root_statement.model.prototype_instance(sym_or_model, id, model, weaver_stackframe) : sym_or_model 
        component(id, prototype, &block)
      end

      # If prototype is nil, we will create a core Component, otherwise inherit attributes,
      # default attribute values, and validators from the prototype
      def component(id, prototype = nil, &block)
        @@logger.debug "component[(#{prototype.nil? ? 'component' : prototype.type}): #{id}]"
        
        return model.add_validation_failure("Do not use hash notation on components") unless id.kind_of?(Symbol)
        
        return if fail_id_redefinition(id, Component)
        
        component_model = @model.components[id]
        if component_model.nil?
          component_model = prototype || root_statement.model.prototype_instance(:component, id, model, weaver_stackframe)
          effective_provisioner = @model.respond_to?(:effective_provisioner) ? @model.effective_provisioner(:artifact, nil) : nil
          unless effective_provisioner.nil?
            if component_model.provider_mixins.empty?
              effective_provisioner.mixin_provider_attributes(component_model)
            end
          end
          @model.components[id] = component_model
        else
          root_statement.model.widen(component_model, prototype) unless prototype.nil?
        end
        
        unless block.nil?
          component = ComponentStatement.new(component_model, self)
          
          begin
            component.instance_eval &block
          rescue ArgumentError => ae
            # For example the statement 'automation' without an ID yields
            # 'wrong number of arguments (0 for 1)'
            component_model.add_validation_failure(" #{ae.message}", weaver_stackframe(ae.backtrace))
            @@logger.warn "Caught #{ae.inspect} processing 'component' statement, backtrace=#{ae.backtrace}"
          end
        end

        return component_model
      end
      
    end

  end
end
