require 'weaver/core/cloud'
require 'weaver/dsl/link_statement'
require 'weaver/dsl/topology_container_statement'
require 'weaver/dsl/provider_container_statement'
require 'weaver/dsl/group_container'
require 'weaver/dsl/user_container'
require 'weaver/dsl/link_container'
require 'weaver/dsl/node_container'
require 'weaver/dsl/service_container_statement'
require 'weaver/dsl/artifact_container'
require 'weaver/dsl/storage_container_statement'

module Weaver
  module Dsl
    # The TopologyStatement implements the processing of the language element 'topology'.
    #
    # Author:: Ed Snible
    # @see Weaver::Core::Topology
    class TopologyStatement < AbstractStatement
      include GroupContainer
      include UserContainer
      include LinkContainer
      include NodeContainer
      include StorageContainerStatement
      include TopologyContainerStatement
      include ServiceContainerStatement
      include ProviderContainerStatement
      include PolicyContainer
      include ArtifactContainer

      # * +id+ Symbol representing the entity, or a Core::Topology which is the entity
      # * +parent_statement+ The enclosing AbstractStatement
      def initialize(id, parent_statement, type = :topology)
        @@logger.debug "topology[id: #{id.kind_of?(Symbol) ? id : id.class} on #{parent_statement.model.type} :#{parent_statement.model.id}]"
        super(id, type, parent_statement)
      end
      
      # Process +cloud"+     
      # * +type+ Symbol which should be <code>:iwd_vsp</code> for IWD Virtual System Patterns
      # * +args+ Hash which may contain <code>:url</code>, <code>:description</code>, and <code>:version</code>
      def cloud(type, args) 
        @@logger.debug "cloud[type: #{type}, args: #{args}]"
        cloud = Cloud.new(type)
        cloud.url = args[:url]
        cloud.description = args[:description]
        cloud.version = args[:version]
        @model.cloud = cloud
        
        # Create a provider, which will be serialized
        if model.providers.empty?
          model.providers[type] = Provider.new(type, model)
          model.providers[type].mixin_provider_attributes(model)
          
          # Legacy processing for IWD
          model.providers[type].host = cloud.url if type == :iwd_vsp
        end
      end
      
      # Process +import+ keyword 
      def import(filespec_with_namespace)
        raise WeaverError.new("'import' statement not allowed here; place it before the topology")
      end
      
      # Process +startup_order+ keyword 
      def startup_order(*nodes)
        @@logger.debug "startup_order #{nodes.inspect}"
        
        model.startup_order = nodes.map do |ele| 
          ele.kind_of?(Array) ? ele.map { |nodeexpr| lookup_node(nodeexpr) } : lookup_node(ele)
        end
        
        model.startup_order.instance_variable_set(:@weaver_origin, weaver_stackframe)
      end
      
      # Process +realizes+ keyword
      def realizes(hsh)
        raise WeaverError.new("'realizes' syntax: source => target") unless hsh.kind_of?(Hash)
        source = hsh.keys.to_a[0]
        raise WeaverError.new("'realizes' source must be a node but it is a #{source.class}") unless source.kind_of?(NodeStatement) || source.kind_of?(FailureEater)
        target = hsh.values.to_a[0]
        raise WeaverError.new("'realizes' target must be a node") unless target.kind_of?(NodeStatement)
        
        @@logger.debug "realizes #{source.model.fully_qualified_id} => #{target.model.fully_qualified_id}"
        source.model.realized_by << target.model
      end
      
      protected
      
      # Given parser input (typically a Symbol or a NodeStatement) determine the Node referred to
      def lookup_node(statement_or_symbol)
        if statement_or_symbol.kind_of?(NodeStatement)
          return statement_or_symbol.model
        end
        
        return model.node_by_id(statement_or_symbol)
      end
      
    end

  end

end
