require 'weaver/dsl/root_statement'
require 'weaver/utils/plugin_loader'

module Weaver
  # == Language 
  # Weaver provides a Domain-Specific Language (DSL) based on the Ruby platform allowing to express the Bill of Material (BOM). The specification of the BOM.
  #
  # \Weaver defines a set of core language constructs that can be used to build a \Weaver specification. These constructs have a pre-defined semantic and most of them can be subtyped for customization.
  # 
  # === {AttributeStatement Properties}
  # === {NodeStatement  Nodes}
  # === {ComponentStatement Components}
  # === {AutomationStatement  Automation}
  # === {LinkStatement Links}
  # === {TopologyStatement  Topology}
  # === {GroupStatement  Group}
  # === {LoadDriverStatement Load Drivers}
  # === {PolicyStatement Policy}
  # === {UserStatement User}
  module Dsl
    # The WeaverParser implements a DSL for .weaver files of any kind (infrastructure,
    # application, or environment).
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    #
    class WeaverParser
      include Weaver::Core

      # @private
      # Cache of the builtins.  (To get the builtins without a file, call parse(nil))
      @@builtins = nil
      
      # @private
      # Cache of .weaver files requested by Weaver plugins before @@builtins are ready
      @@requested_extensions = []
      # @private
      @@loaded_extensions = []
      
      # @private
      # A Hash that no matter what you try to dereference returns itself.  Used to keep
      # config[:typo][:key] from blowing up when parsing invalid .weaver files.
      PROMISCIOUS_HASH_OF_NILS = {}
      # @private
      def PROMISCIOUS_HASH_OF_NILS.[](key)
        return PROMISCIOUS_HASH_OF_NILS
      end
      
      # The current default directory for reading imported/included/refined .weaver files
      attr_reader :dir

      # The definitions of statement IDs
      attr_reader :definitions
      
      # @return [Hash<String, Object>] config.yml with overrides already processed
      attr_reader :config
      
      # @param [Hash] config configuration values that Weaver files can link to attributes
      def initialize(config = {})
        @definitions = {}
        @config = config.clone
#        @config.default = PROMISCIOUS_HASH_OF_NILS # Don't crash if config[] is passed a bogus key
        @config.instance_exec {
          def [](key)
            return super if self.key?(key)
            @@logger.warn "Unknown configuration parameter config[#{key.inspect}], keys=#{self.keys}, caller=#{caller}"
            PROMISCIOUS_HASH_OF_NILS
          end
        }
      end

      # Parses and processes the file +weaver_file+.
      #
      # * *Args*    :
      #   - +weaver_file+ -> any .weaver file (as filename String or File object)
      #   - +context+ -> any {AbstractStatement} or nil for top-level .weaver files
      # * *Returns* :
      #   - An Array of Weaver objects that can be used for validation and deployment.
      # * *Raises* :
      #   - +WeaverError+ -> if an error occurs while resolving all the files.
      #
      def parse(weaver_file, context = nil)
        parse_using_load_path(weaver_file, nil, context)
      end
      
      # Like parse(), but allows 'weaver_file' to be relative to any entry in $LOAD_PATH
      def parse_using_load_path(weaver_file, dir, context = nil)
        weaver_file = file(weaver_file, dir) if weaver_file.kind_of?(String)
        unless weaver_file.nil?
          @@logger.info "Processing file: '#{weaver_file.path}'"
          contents = weaver_file.read()
          weaver_file.close()
          filename = weaver_file.path
        else
          contents = ''
          filename = ''
        end
        
        is_top_level = context.nil?
        context = create_root_statement(filename) if context.nil?
        context.dir = File.dirname(filename)
        
        begin
          context.instance_eval(contents, filename, 1)
        rescue ArgumentError => ae
          # For example the statement 'application' without an ID yields
          # 'wrong number of arguments (0 for 1)'
          context.model.add_validation_failure(ae.message, weaver_stackframe(filename, ae.backtrace))
          @@logger.warn "Caught #{ae.inspect} parsing; backtrace=#{ae.backtrace}"
        rescue SyntaxError => se
          context.model.add_validation_failure(se.message, weaver_stackframe(filename, se.backtrace))
          @@logger.warn "Caught #{se.inspect} parsing; backtrace=#{se.backtrace}"
        rescue WeaverError => we
          context.model.add_validation_failure(we.message, weaver_stackframe(filename, we.backtrace))
          @@logger.warn "Caught #{we.inspect} parsing; backtrace=#{we.backtrace}"
        rescue NoMethodError => nme
          context.model.add_validation_failure(nme.message, weaver_stackframe(filename, nme.backtrace))
          @@logger.warn "Caught #{nme.inspect} parsing; backtrace=#{nme.backtrace}"
        rescue TypeError => te
          context.model.add_validation_failure(te.message, weaver_stackframe(filename, te.backtrace))
          @@logger.warn "Caught #{te.inspect} parsing; backtrace=#{te.backtrace}"
        rescue NameError => ne
          context.model.add_validation_failure(ne.message, weaver_stackframe(filename, ne.backtrace))
          @@logger.warn "Caught #{ne.inspect} parsing; backtrace=#{ne.backtrace}"
        rescue Exception => ex
          context.model.add_validation_failure(ex.message, weaver_stackframe(filename, ex.backtrace))
          @@logger.error "Caught #{ex.inspect} parsing; backtrace=#{ex.backtrace}"
        end

        @@logger.debug "Finished processing file: '#{weaver_file.path}'" unless weaver_file.nil?
        
        if is_top_level
          # Create pseudo-validation-failure for config[expression] in the .weaver file that didn't work
          context.model.flatten_children.select { |child| child.kind_of?(Attribute) && child.unresolved_value.equal?(PROMISCIOUS_HASH_OF_NILS) }.each do |attr|
            attr.add_validation_failure("Unknown config key")
            attr.value = nil
          end
          
          parse_encrypted_passwords(context.model)
        end
        
        return context.model
      end

      def file(filespec, dir)
        return File.new(filespec) if dir.nil? && File.exists?(filespec)
        return File.new(dir + File::SEPARATOR + filespec) if dir != nil && File.exists?(dir + File::SEPARATOR + filespec)
        $LOAD_PATH.map {|item| File.expand_path(item) }.uniq.each do |path|
          return File.new(path + File::SEPARATOR + filespec) if File.exists?(path + File::SEPARATOR + filespec)
        end
        raise Errno::ENOENT.new(filespec) # Explicitly fail, so we don't lose track of where/how the filename was mis-specified
#        return File.new(filespec) # doesn't exist, though
      end
      
      # Allow Weaver plug-ins to supply built-in types and validators
      def self.extend_builtins(filename)
        if @@builtins.nil?
          @@logger.info "Queuing request to load #{filename.inspect}"
          @@requested_extensions << filename
          return
        end
        
        if @@loaded_extensions.include?(filename)
          @@logger.info "Weaver extension #{filename.inspect} is already loaded"
          return
        end

        context = RootStatement.new(filename, WeaverParser.new())
        context.model = @@builtins
        
        begin
          orig_origin = context.model.weaver_origin
          context.model.weaver_origin = "#{filename}:1"
          context.instance_eval(File.new(filename).read(), filename, 1)
          context.model.weaver_origin = orig_origin
          @@loaded_extensions << filename
        rescue Weaver::Core::WeaverError => we
          @@logger.warn "Couldn't parse #{filename}, exception #{we}"
        end
      end
      
      private
      
      # This is a rough sketch of code to look up the .weaver filename and
      # line number by inspecting the call stack.
      # * *Args*    :
      #   - +filename+ -> filename to look for 
      #   - +bt+ -> Array representing a stack frame.  Pass Exception#backtrace or Kernel#caller
      # TODO Refactor and combine with AbstractStatement.weaver_stackframe
      def weaver_stackframe(filename, bt)
        pattern = "#{filename}:" # Filename followed by colon as it appears in a Ruby stackframe
        line_in_backtrace = bt.find {|stackframe| stackframe.include? pattern }
          
        # If we don't find the file we expected look for any .weaver: file, e.g. builtins.weaver
        line_in_backtrace = bt.find {|stackframe| stackframe.include? ".weaver:"} if line_in_backtrace.nil?
          
        return nil if line_in_backtrace.nil?
        md = /(?<sourcefile>.*):(?<lineno>[0-9]*):in `(.*parse|\(eval\))'/.match(line_in_backtrace)
        return md.nil? ? line_in_backtrace : "#{md[1]}:#{md[2].to_i}"
      end
      
      # Go through all attributes that are direct or indirect children of +entity+ and replace 'decrypt(...)' with an EncryptedValue
      def parse_encrypted_passwords(entity)
        entity.flatten_attributes.select do |attr| 
          attr.unresolved_value.kind_of?(String) && Weaver::Utils::ValueDecryptor::DECRYPT_REGEXP.match(attr.unresolved_value)
        end.each do |attr|
          attr.value = EncryptedValue.new(Weaver::Utils::ValueDecryptor.base64_from_expression(attr.value))
        end
      end
      
      def create_root_statement(filename)
        context = RootStatement.new(filename, self)
        context.model.weaver_origin = "#{filename}:1"
        
        unless @@builtins.nil?
          context.model.instance_exec(@@builtins) do |b|
            @named_validators.merge!(b.named_validators)
            @rules.merge!(b.rules)
            @weaver_extended_types.merge!(b.instance_exec { @weaver_extended_types })
          end
        else
          full_path = File.expand_path(File.dirname(__FILE__) + '/../builtins.weaver')
          @@logger.info "Loading #{full_path.inspect}"          
          begin
            context.instance_eval(File.new(full_path).read(), full_path, 1)
          rescue Weaver::Core::WeaverError => we
            @@logger.warn "Couldn't parse #{full_path}, exception #{we}"
          end
          
          # Set @@builtins so it can be used by the plugins, which MAY call back via 
          # #extend_builtins() to add new builtin types
          @@builtins = context.model
          
          # Process any extensions requested before @@builtins was available
          @@requested_extensions.each { |filename| WeaverParser::extend_builtins(filename) }
          
          # load all plugins in the system.  The plugins MAY contribute other builtins
          Weaver::Utils::PluginLoader.new(
            [File.join("weaver", "plugins", "deployers")],
            [File.expand_path(File.join("deployers", "plugins"), File.dirname(__FILE__))]
          ).load_plugins
          
          # Don't let context.model hold the real @@builtins; we don't want to pollute it
          # with validators and types from regular .weaver files
          context.model = @@builtins.clone
          
        end
        
        return context
      end

      
    end

  end # module Dsl
end # module weaver
