require 'weaver/core/attribute'
require 'weaver/core/logging'
require 'weaver/core/weaver_error'
require 'weaver/transforms/unroll'
require 'weaver/transforms/duplicate'

module Weaver
  module Transforms

    # Implements merge of current state model with desired state model
    #
    # Author:: Michael Kalantar
    class Merge
      include Weaver::Core
      
      # Create an entity by merging two entities
      # * +desired_state+ a Base, typically a Topology or Root, to be merged
      def self.merge(desired_state, current_state)
        raise Weaver::Core::WeaverError.new("Can't merge nil") if desired_state.nil?
        # current_state can be nil

        merger = Merge.new(desired_state, current_state)
        return merger.merged_entity
      end # def self.unroll

      # The actual merged value
      def merged_entity
        return @entity
      end # def merged_entity

      protected

      # * +desired_state+ a {Base}, typically a {Topology} or {Root}, to be unrolled
      def initialize(desired_state, current_state = nil)
        out = []
        @entity = Duplicate.duplicate(desired_state)
        unless current_state.nil?
          merge(current_state)
        end # unless current_state.nil?
      end # def initialize

      # Merge a model of a currently deployed environment into a new environment request model (@entity)
      # Currently supports the following limited merging 
      #    nodes (and their attributes) [not storage, components, services, automations...]
      #    environment level attributes 
      def merge (current_instance)
        
        merge_nodes(current_instance, @entity)
        merge_environment_attributes(current_instance, @entity)
          
        @@logger.debug "Updated request is: #{@entity.to_weaver}"
      end # def merge
      
      protected
      
      def merge_nodes (current_env, request_env)
        request_nodes = request_env.flatten_children.select { |n| n.is_a?(Weaver::Core::Node) }
        current_env.flatten_children.select {|n| n.is_a?(Weaver::Core::Node) and !Weaver::Transforms::Unroll.template?(n)}.each do |node|
          request_node = request_nodes.find { |rn| node.fully_qualified_id == rn.fully_qualified_id }
          if request_node.nil?
            # the node is in the current model but not the request model
            # copy it into the request model 
            # ... but add a :to_delete tag 
          else
            # we've matched current_node with a request_node
            # if the request node is a template (--> current node is a template), re-map to the first node in the list
            if Weaver::Transforms::Unroll.template?(request_node) then
              request_node = Weaver::Transforms::Unroll.group(request_node).members.values.find {|m| !Weaver::Transforms::Unroll.template?(m)}
            end # if Weaver::Transforms::Unroll.template?(request_node) then
            @@logger.debug "Matched instance node #{node.id.inspect} to request_node #{request_node.id.inspect}"
            
            # merge attributes from current_node to request_node
            merge_node_attributes(node, request_node)
          end # if request_node.nil?
        end # current_instance.flatten_children.select{}.each do
      end # def merge_nodes
      
      # Create a copy of a node (from the currently deployed environment) to the request.
      # Such nodes are nodes should be deleted and should be marked as such. 
      def duplicate_node (current_node)
        current_scope = current_node.scope
        request_scope = @entity.flatten_children.find{|c| c.fully_qualified_id == current_scope.fully_qualified_id }
        request_node = Weaver::Transforms::Duplicate.duplicate(current_node)
        request_scope.nodes[request_node.id] = request_node
        request_node.tags << :to_delete
      end # def duplicate_node
      
      # Merge (most) attributes from the deployed node to the request node unless already defined on the request node
      def merge_node_attributes (current_node, request_node)
        # hostname and ip_addresses are not defined in the request_node but should be
        request_node.hostname = current_node.hostname
        request_node.ip_address = current_node.ip_address
        if current_node.nics.include?(:private_ip) then
          request_node.private_ip.ip_address = current_node.private_ip.ip_address
        end
        
        # more attributes defined in current_node and not request_node
        [:config_status, :vm_status].each do |id|
          unless current_node.amd[id].nil?
            request_node.amd[id] = Weaver::Transforms::Duplicate.duplicate(current_node.amd[id])
          end
        end # [].each do
        
        # copy everything else from current_node
        #    UNLESS the attribute is already defined on the request_node or
        #            the attribute is in a black list
        dont_copy_list = [:user_data, :user_data_encoded]
          
        current_node.amd.each do |id, attr|
          unless request_node.amd.include?(id) or dont_copy_list.include?(id) then
            request_node.amd[id] = Weaver::Transforms::Duplicate.duplicate(current_node.amd[id])
          end 
        end # current_node.amd.each do
        
        # copy value of multiplicity index
        # request_node.amd[:multiplicity_index].value = node.amd[:multiplicity_index].value
      end # def merge_node_attributes
      
      def merge_environment_attributes (current_env, request_env)
        top = request_env.topologies.values.first
        top.amd[:config_status] = Weaver::Core::Attribute.new(:config_status, top)
        top.amd[:config_status].value = current_env.topologies.values.first.config_status
        top.amd[:deployment] = Weaver::Core::Attribute.new(:deployment, top)
        top.amd[:deployment].value = current_env.topologies.values.first.deployment
      end # def merge_environment_attributes
      
      end # class Merge
  
    end # module Transforms
  end # module Weaver
