require 'yaml'
require 'date'

module Weaver
  module Utils
    # This class writes a yaml file of a Base class's mutable properties
    class YamlPropertyWriter
      include Weaver::Core
      # Root of the Weaver document to be checked
      attr_reader :topology
      def initialize(topology)
        @topology = topology
      end

      def write_yaml(filename, write_unnamed=false)
        unless @topology.nil?
          properties = collect_properties(@topology, write_unnamed)
#         puts ">>>>>>>> properties size is #{properties.size}"
          @@logger.info("Writing infrastructure yaml file to #{filename}...")
          File.open(filename, 'w') do |f|
            prefix = "# Generated automatically by weaver on #{DateTime.now.strftime('%FT%R')}"
            f.write("#\n" + prefix + "\n#\n" + properties.to_yaml)
          end
        end
      end

      # Returns a Hash of all attributes, with attributes of children nested under the children
      # RECURSIVE
      def collect_properties(base, write_unnamed=false, parent = nil)
        # add the 'name' property for the top level
        properties = Hash.new
          
        amd = base.amd
        # for each attribute, add a Hash of the attribute id to its value
        # only collect unnamed attributes if 'write_unnamed' is true.
        # always collect the :name property of the top level base that's passed to the method
        amd.each do |id, attr|
          if(attr.mutable && !(attr.value.kind_of?(LateBinding)))
            if ((write_unnamed || (!write_unnamed && !attr.name.nil?)) || (parent.nil? && attr.id == :name))
              #             puts(">>>>> ATTRIBUTE #{attr.id}, #{attr.name.nil? ? "nil name" : attr.name}, value = #{attr.value}, value class is #{attr.value.class}, #{attr.mutable}")
              value = get_value(attr)
              if (value.nil? || (value.kind_of?(String) && value.length == 0))
                properties[attr.id.to_s] = nil
              else
                properties[attr.id.to_s] = value
              end
            end
          end
        end

        # add a Hash of a base child's id to its properties
        base.child_entities.each_value do |child|
          if (child.kind_of?(Base) && !child.kind_of?(Attribute))
            unless properties.has_key?(child.id.to_s)
              #             puts ">>>>>>>>>>>>> CHILD #{child.id.to_s} class #{child.class}"
              child_properties = collect_properties(child, write_unnamed, base)
              # only write the hash if the child has properties
              unless child_properties.empty?
                properties[child.id.to_s] = child_properties
              end

            end
          end
        end

        return properties
      end
      
      # Changes the value of each property to a reference
      # RECURSIVE
      def set_values_to_references(base, accum_keys = nil, write_unnamed=false, parent = nil)
        accum_keys = [] if accum_keys.nil?
        properties = Hash.new

        # for each attribute, add a Hash of the attribute id to its value
        # only collect unnamed attributes if 'write_unnamed' is true.
        # always collect the :name property of the top level base that's passed to the method
        amd = base.amd
        amd.each do |id, attr|
          if(attr.mutable && !(attr.value.kind_of?(LateBinding)))
            if ((write_unnamed || (!write_unnamed && !attr.name.nil?)) || (parent.nil? && attr.id == :name))
              #              puts(">>>>> ATTRIBUTE #{attr.id}, #{attr.name.nil? ? "nil name" : attr.name}, #{attr.value}, #{attr.mutable}")

              if (attr.id == :chef_node_attr)
                #              attr.value = UnquotedString.new("late_binding { Weaver::Chef::ChefAutomationHandler.interpolate_and_convert_chef_node_attr_to_s(" + properties_string(accum_keys, attr.id.to_s) + ", self) }")
                update_chef_node_validator(attr)
                #              else
                #                attr.value = UnquotedString.new(properties_string(accum_keys, attr.id.to_s))
              end
              # for the top-level "name" attribute, append the DateTime. This is important for IWD because IWD uses the "name" 
              # attribute as the virtual system instance name, and a virtual system instance name must be unique
              if (parent.nil? && attr.id == :name)
                attr.value = UnquotedString.new(properties_string(accum_keys, attr.id.to_s) + " + \"_\#{DateTime.now.strftime('%Y%m%dT%H%M%S')}\"")
              else
                attr.value = UnquotedString.new(properties_string(accum_keys, attr.id.to_s))
              end
              if (attr.value.nil? || (attr.value.kind_of?(String) && attr.value.length == 0))
                properties[attr.id.to_s] = nil
              else
                properties[attr.id.to_s] = attr.value
              end
            end
          end
        end

        # add a Hash of a base child's id to its properties
        base.child_entities.each_value do |child|
          if (child.kind_of?(Base) && !child.kind_of?(Attribute))
            unless properties.has_key?(child.id.to_s)
              #             puts ">>>>>>>>>>>>> CHILD #{child.id.to_s} class #{child.class}"
              accum_keys.push(child.id.to_s)
              child_properties = set_values_to_references(child, accum_keys, write_unnamed, base)
              accum_keys.pop()
              # only write the hash if the child has properties
              unless child_properties.empty?
                properties[child.id.to_s] = child_properties
              end

            end
          end
        end

        return properties
      end
      
      def properties_string(keys, id)
        str = "properties"
        keys.each do |key|
          str = str + "['#{key}']"
        end
        str = str + "['#{id}']"
      end
      
      def get_value(attr)
        if attr.value != nil
          v = attr.value
        else
          v = attr.default
        end
        return v
      end
      
      def update_chef_node_validator(attr)
        delete_validator(attr, :value_type)
      end

      def delete_validator(attr, sym)
        val_to_delete = nil
        unless attr.validators.nil?
          attr.validators.each  do | val |
            if val[0] == sym
            val_to_delete = val
            end
          end
          attr.validators.delete(val_to_delete) unless val_to_delete.nil?
        end
      end

      # @private
      # A String that *doesn't* have quotes when inspected
      class UnquotedString < String # :nodoc:
        def initialize(s)
          super
        end

        def inspect
          return to_s
        end
      end

    end
  end
end